#! /usr/bin/env python
# -*- coding: utf-8 -*-
import sys, time
from .. import es

SubjectIndexs = {
	"count": 0,
	"play": 1,
	"replay": 2,
	"conv": 3
}

WorksIndexs = {
	"play": 0,
	"like": 1,
	"shareF": 2,
	"shareI": 3,
	"replay": 4,
	"longTime": 5,
	"subAdd": 6,
	"playTimeAvg": 7,
	"conv": 8,
	"duration": 9,
	"size": 10
}

DurationSections = [0, 30, 60, 180, 300, 600, 900, 1200, 1800, 2400, 3600, 7200]


def createWorksArgsList(worksPersonas, aids=[], blackVids=[]):
	subjectDict, worksDict = createWorksDict(aids, blackVids)
	addBehaviors(subjectDict, worksDict, worksPersonas)
	return subjectDict, worksDict

def updateNewestActions(dau, worksPersonas):
	addActions(dau, worksPersonas, es.searchPersonas(worksPersonas.lastPersonasTime))

def addActions(dau, worksPersonas, personas):
	for index, hit in enumerate(personas):
		'''
		if index % 100 == 0:
			sys.stdout.write(str(index) + '\r')
			sys.stdout.flush()
		'''
		source = hit['_source']
		uid = source['uid']
		body = source['body']
		actions = body.split('\n')
		for action in actions:
			items = action.split('\t')
			if len(items) > 2 and items[2].isdigit():
				cTime = long(items[0])
				dau.addAction(uid, cTime)
				worksPersonas.addAction(uid, cTime, items[1], items[2], items[3:] if len(items) > 3 else [])
		if source.has_key('create_time'):
			worksPersonas.updateLastPersonasTime(source['create_time'])

def getDurationSectionsAvg(worksDict):
	length = len(DurationSections)
	durationSectionsInfo = []
	for index in range(length + 1):
		durationSectionsInfo.append([0, 0, 0])
	for works in worksDict.values():
		durationSectionsInfo[works['durationIndex']][0] += works['playTime']
		durationSectionsInfo[works['durationIndex']][1] += works['timeCount']
		durationSectionsInfo[works['durationIndex']][2] += 1
	durationSectionsAvg = []
	durationSectionsAvgCount = []
	for index in range(length + 1):
		timeCount = durationSectionsInfo[index][1]
		durationSectionsAvg.append(0 if timeCount == 0 else durationSectionsInfo[index][0] * 1.0 / timeCount)
		worksCount = durationSectionsInfo[index][2]
		durationSectionsAvgCount.append(0 if worksCount == 0 else timeCount)# * 1.0 / worksCount)
	return durationSectionsAvg, durationSectionsAvgCount

def addBehaviors(subjectDict, worksDict, worksPersonas):
	for behavior in worksPersonas.behaviors.values():
		vid = behavior['vid']
		if worksDict.has_key(vid):
			works = worksDict[vid]
			addBehavior(subjectDict[works['aid']], works, behavior)
	for subject in subjectDict.values():
		if subject['show'] > 0:
			subject['statis'][SubjectIndexs['conv']] = subject['showPlay'] * 1.0 / subject['show']

	durationSectionsAvg, durationSectionsAvgCount = getDurationSectionsAvg(worksDict)

	for works in worksDict.values():
		if works['playTime'] > 0 and works['timeCount'] > 0:
			works['statis'][WorksIndexs['playTimeAvg']] = works['playTime'] * 1.0 / works['timeCount'] / durationSectionsAvg[works['durationIndex']]
		if works['showPlay'] > 0 and works['show'] > 0:
			works['statis'][WorksIndexs['conv']] = works['showPlay'] * 1.0 / works['show']
		works['durationAvgPlay'] = durationSectionsAvgCount[works['durationIndex']]

def addBehavior(subject, works, behavior):
	if behavior['play'] > 0:
		works['statis'][WorksIndexs['play']] += 1
		subject['statis'][SubjectIndexs['play']] += 1
		if behavior['play'] > 1:
			works['statis'][WorksIndexs['replay']] += 1
			subject['statis'][SubjectIndexs['replay']] += 1
	if behavior['like']:
		works['statis'][WorksIndexs['like']] += 1
	if behavior['shareF'] > 0:
		works['statis'][WorksIndexs['shareF']] += 1
	if behavior['shareI'] > 0:
		works['statis'][WorksIndexs['shareI']] += 1

	if behavior['maxPlayTime'] > min(60, works['statis'][WorksIndexs['duration']] * 0.8):
		works['statis'][WorksIndexs['longTime']] += 1

	works['statis'][WorksIndexs['subAdd']] += behavior['subAdd']
	works['playTime'] += behavior['playTime']
	works['timeCount'] += behavior['timeCount']

	works['recShow'] += 1 if behavior['recShow'] > 0 else 0
	works['recShowPlay'] += 1 if behavior['recShowPlay'] > 0 else 0

	works['show'] += behavior['show']
	subject['show'] += behavior['show']
	showPlay = min(behavior['show'], behavior['showPlay'])
	works['showPlay'] += showPlay
	subject['showPlay'] += showPlay


def createWorksDict(aids=[], blackVids=[]):
	subjectDict = {}
	worksDict = {}
	for hit in es.scrollSearch('works_video_read', 'video', {"size":1000}):
		worksInfo = hit['_source']
		if not worksInfo.has_key('vid') or not worksInfo.has_key('aid') or (worksInfo.has_key('state') and worksInfo['state'] == 1) or (worksInfo.has_key('authorstate') and worksInfo['authorstate'] == 1):
			continue
		if not worksInfo.has_key('publishtime') or worksInfo['publishtime'] == None:
			continue
		publishTime = int(time.time() - es.esTimeToTime(worksInfo['publishtime'], True)) / 3600
		if publishTime > 336:
			continue
		aid = worksInfo['aid']
		vid = worksInfo['vid']
		if (len(aids) != 0 and not aid in aids) or vid in blackVids:
			continue
		if subjectDict.has_key(aid):
			subject = subjectDict[aid]
			subject['statis'][SubjectIndexs['count']] += 1
		else:
			subject = subjectDict[aid] = {'show': 0, 'showPlay': 0, 'statis': [1, 0, 0, 0]}
		duration = worksInfo['duration'] if worksInfo.has_key('duration') else 0
		size = worksInfo['filesize'] if worksInfo.has_key('filesize') else 0
		durationIndex = 12
		for index, value in enumerate(DurationSections):
			if duration < value:
				durationIndex = index
				break
		worksDict[vid] = {'title': worksInfo['title'], 'aid': aid, 'publishTime': publishTime, 'playCount': worksInfo['playcount'] if worksInfo.has_key('playcount') else 0, 'playTime': 0, 'timeCount': 0, 'show': 0, 'showPlay': 0, 'recShow': 0, 'recShowPlay': 0, 'durationIndex': durationIndex, 'durationAvgPlay': 0, 'statis': [0, 0, 0, 0, 0, 0, 0, 0, 0, duration, size]}

	return subjectDict, worksDict

'''

def writeWorksHot(esclient, index, worksHots):
	esclient.delete_by_query(index=index, doc_type='video', body={"query":{"match_all":{}}})
	doc = []
	for item in worksHots:
		doc.append({"index":{"_id":item[0]}})
		doc.append({"vid":item[0],"hot":item[1],"worksHot":item[2],"subjectHot":item[3],"timeDecay":item[4],"durationRate":item[5],"duration":item[6],"size":item[7],"publishTime":item[8],"playUserCount":item[9],"show":item[17],"showPlay":0 if item[17] == 0 else item[18] * 1.0 / item[17]})
	esclient.bulk(index=index, doc_type='video', body=doc)

def writeWorksBase(esclient, index, worksDict):


def createWorksList(isTest):
	subjectDict, worksDict = dealWithWorks(searchWorks(isTest))

	worksDict = updateHistoryWorks(isTest, worksDict)
	dau = getHistoryDau(isTest)
	
	worksDict, dau = dealWithPersonas(searchPersonas(isTest), worksDict, dau)

	return subjectDict, worksDict, dau

def updateHistoryWorks(isTest, worksDict):
	historyData = searchStatisticsData(isTest, 'statistics_works')
	for hit in historyData:
		vid = hit['_id']
		if worksDict.has_key(vid):
			worksDict[vid].update(hit['_source'])
		#TODO::后续要加上删除
	return worksDict
		
def getHistoryDau(isTest):
	dau = {}
	historyData = searchStatisticsData(isTest, 'statistics_dau')
	for hit in historyData:
		dau[hit['_id']] = hit['_source']['count']
	return dau

def toWriteWorksPersonas(action):
	


def dealWithWorks(works):
	subjectDict = {}
	worksDict = {}
	for hit in works:
		worksInfo = hit['_source']
		if not worksInfo.has_key('vid') or not worksInfo.has_key('aid'):
			continue
		if not subjectDict.has_key(worksInfo['aid']):
			subjectDict[worksInfo['aid']] = Subject(worksInfo['aid'])
		subject = subjectDict[worksInfo['aid']]
		works = worksDict[worksInfo['vid']] = Works(subject, worksInfo['vid'], worksInfo['duration'], 0, worksInfo['publishtime'])
		subject.addWorks(works)

	return subjectDict, worksDict

def dealWithPersonas(personas, worksDict, dau):
	now = time.localtime(time.time())
	zeroTime = time.mktime((now.tm_year, now.tm_mon, now.tm_mday, 0, 0, 0, 0, 0, 0))
	uids = [[], [], [], [], [], [], []]
	for index, hit in enumerate(personas):
		if index % 100 == 0:
			sys.stdout.write(str(index) + '\r')
    		sys.stdout.flush()
		uid = hit['_source']['uid']
		body = hit['_source']['body']
		actions = body.split('\n')
		for action in actions:
			items = action.split('\t')
			if len(items) > 2 and items[2].isdigit():

				calActive(uids, zeroTime, int(items[0]), uid)
				if items[2] in Works.DealFunc.keys():
					if items[2] == '7':
						if items[1] == 'pages/common/videoplay/videoplay' and len(items) >= 5 and items[3] == '1':
							vid = items[4]
							args = items[5:]
						else:
							continue
					elif items[2] == '10' or items[2] == '11':
						vid = items[4] if len(items) > 4 else ''
						if vid == '':
							continue
					else:
						vid = items[3]
						args = items[4:]
					if worksDict.has_key(vid):
						worksDict[vid].addAction(items[2], uid, args)

	dau = []
	for dayUids in  uids:
		dau.append(len(dayUids))
	return worksDict, dau


def calActive(uids, zeroTime, actionTime, uid):
	for day in range(7):
		if actionTime >= (zeroTime - oneDay * day) and not uid in uids[day]:
			uids[day].append(uid)
			return

def scrollSearch(isTest, index, doc_type, body):
	esclient = testEsclient if isTest else relEsclient
	hits = []
	result = esclient.search(index=index, doc_type=doc_type, scroll='1m', body=body)
	while len(result['hits']['hits']) != 0:
		hits.extend(result['hits']['hits'])
		scroll_id = result['_scroll_id']
		result = esclient.scroll(scroll_id=scroll_id, scroll='1m')
	return hits


def searchStatisticsData(isTest, index)
	body = {
		"size": 1000
	}
	return scrollSearch(isTest, index, 'test' if isTest else 'rel', body)


def searchWorks(isTest):
	body = {
		"size": 1000
	}
	return scrollSearch(isTest, 'works_video_read', 'video', body)




def write(filename, itemDict):
	doubleArray = []
	for item in itemDict.values():
		doubleArray.append(item.getArgs())
	with open(filename, 'w') as f:
		data = []
		for itemArray in doubleArray:
			data.append('\t'.join(map(str, itemArray)))
		f.write('\n'.join(data))
'''





#! /usr/bin/env python
# -*- coding: utf-8 -*-

import time, json
from .. import es

class BftWorks(object):

	RiseThreshold = 0.3

	PoolLevelMax = 3

	ShowMax = {0: 50, 1: 100, 2: 500}

	Weights = {0: {"like": 1.5, "share": 1.5, "longPlay": 1, "play": 1}
	, 1: {"like": 2, "share": 2, "longPlay": 1, "play": 1}
	, 2: {"like": 2, "share": 3, "longPlay": 1, "play": 1}}

	#1 正常 2 晋级 3 待定 4 淘汰
	Status = {'Normal': 1, 'Success': 2, 'Await': 3, 'Fail': 4}

	def __init__(self, vid):
		self.vid = vid
		self.records = []
		self.score = 0
		self.enterPool(0)

	def updateAction(self, action):
		if self.status in [BftWorks.Status['Success'], BftWorks.Status['Fail']]:
			return

		actionId = action['action']
		uid = action['uid']
		if actionId == 3:
			self.uidSetDict['show'].add(uid)
		elif actionId == 16:
			self.uidSetDict['like'].add(uid)
		elif actionId == 12 or actionId == 13:
			self.uidSetDict['share'].add(uid)
		elif actionId == 2:
			self.uidSetDict['longPlay'].add(uid)
		elif actionId == 5:
			self.uidSetDict['play'].add(uid)

	def updateStatus(self):
		if self.status in [BftWorks.Status['Success'], BftWorks.Status['Fail']]:
			return

		pool = self.getPool()
		showMax = BftWorks.ShowMax[pool]
		weight = BftWorks.Weights[pool]

		show = self.getCount('show')
		score = 0
		totalOne = 0
		for key in weight.keys():
			score += weight[key] * self.getCount(key)
			totalOne += weight[key]
		
		if score >= max(showMax, show) * totalOne * BftWorks.RiseThreshold:
			self.rise()
		elif show >= showMax:
			self.status = BftWorks.Status['Await']

		self.score = score

	def updateHistoryActions(self, endTime):
		if self.status in [BftWorks.Status['Success'], BftWorks.Status['Fail']]:
			return

		body = {"query": {"bool": {"must": [{"range": {"time": {"gte": self.records[0]['time'] * 1000, "lt": endTime * 1000}}}, {"term": {"vid": self.vid}}, {"terms": {"action": [2, 3, 5, 12, 13, 16]}}]}}, "size": 0, "aggs": {}}
		body['aggs']['uid_diff'] = {"terms": {"field": "action", "size": 6}, "aggs": {"count": {"cardinality": {"field": "uid"}}}}
		aggs = es.searchStatisticsAggs('works_actions', body)
		
		counts = {2: 0, 3: 0, 5: 0, 12: 0, 13: 0, 16: 0}
		for bucket in aggs['uid_diff']['buckets']:
			counts[bucket['key']] = bucket['count']['value']

		self.historyData['show'] = counts[3]
		self.historyData['like'] = counts[16]
		self.historyData['share'] = counts[12] + counts[13]
		self.historyData['longPlay'] = counts[2]
		self.historyData['play'] = counts[5]

		self.updateStatus()

	def isResurgence(self):
		return self.records[0]['riseType'] == 3

	def resurgence(self):
		self.enterPool(self.getPool(), 3)

	#1 正常晋级（默认） 2 候补补齐
	def rise(self, riseType=1):
		pool = self.getPool()
		if pool >= BftWorks.PoolLevelMax - 1:
			self.status = BftWorks.Status['Success']
		else:
			self.enterPool(pool + 1, riseType)

	def fail(self):
		self.status = BftWorks.Status['Fail']

	#1 正常晋级（默认） 2 候补补齐  3 复活进入
	def enterPool(self, newPool, riseType=1):
		self.records[0]['score'] = self.score
		self.records.insert(0, {'pool': newPool, 'time': int(time.time()), 'riseType': riseType, 'score': 0})

		self.score = 0
		self.status = BftWorks.Status['Normal']
		self.historyData = {'show': 0, 'like': 0, 'share': 0, 'longPlay': 0, 'play': 0}
		self.uidSetDict = {'show': set(), 'like': set(), 'share': set(), 'longPlay': set(), 'play': set()}

	def getPool(self):
		return self.records[0]['pool']

	def getCount(self, key):
		return self.historyData[key] + len(self.uidSetDict[key])

	def toSaveEsData(self):
		data = {"vid": self.vid, 'score': self.score, 'show': self.getCount('show')}
		for key in BftWorks.Weights[self.pool].keys():
			data[key] = self.getCount(key)
		return data

	def syncEsData(self, source):
		self.pool = source["pool"]
		self.timeEnter = source["timeEnter"]
	


	


#! /usr/bin/env python
# -*- coding: utf-8 -*-

#import hot, es, dataDispose, itemCF, test, time, dataSave, filterPool, dataCenter
import es, dataCenter

def startUpdateHot(isTest):
	es.setEnv(isTest)
	hot.startUpdate(isTest, 'config.ini')

def disposeData(isTest):
	es.setEnv(isTest)
	dataDispose.cycleDisposePersonas()

def cfItem(isTest):
	es.setEnv(isTest)
	itemCF.cf()

def updateHistoryActions(isTest):
	es.setEnv(isTest)
	dataDispose.updateHistoryActions()

def saveData(isTest):
	es.setEnv(isTest)
	#print(es.getEsclient('statistics_' + 'works_kartuns' + '_write'))
	dataSave.cycleSave()

def testFunc(isTest):
	es.setEnv(isTest)
	dataCenter.toHive()
	#test.test()
	#filterPool.updateWorks()
	#print(es.searchDbConfig())
	#print('delete post works record: ' + ('success ' if dataDispose.deletePastWorksRecord() else 'fail '))
	#dataDispose.deletePastWorksRecord()
	#test.test()
	'''
	while True:
		es.setEnv(isTest)
		#es.updateRedisWorksHots({'f7106cef64660d0e148d98d77290ff92': 85295.45051608284})
		print(es.searchDbConfig())
		time.sleep(2)
	'''
	



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import gc, time, threading
from .. import es

class Works(object):

	DurationSections = [0, 30, 60, 180, 300, 600, 900, 1200, 1800, 2400, 3600, 7200]

	def __init__(self, vid, aid, title, publishTime, duration):
		self.vid = vid
		self.aid = aid
		self.title = title
		self.publishTime = int(time.time() - es.esTimeToTime(publishTime, True)) / 3600
		durationIndex = 7200
		for value in Works.DurationSections:
			if duration < value:
				durationIndex = value
				break
		self.durationIndex = durationIndex
		self.startTime = -1
		self.endTime = 0
		self.play = 0
		self.shareF = 0
		self.shareI = 0
		self.replay = 0
		self.like = 0
		self.sub = 0
		self.longPlay = 0
		self.timeAvg = 0
		self.score = 0
		self.ratio = 0
		self.resShow = 0
		self.resShowPlay = 0

		self.playUids = []
		self.shareFUids = set()
		self.shareIUids = set()
		self.replayUids = set()
		self.timeTotal = 0
		self.timeCount = 0

	def updateAction(self, action):
		actionId = action['action']
		uid = action['uid']
		if actionId == 1:
			self.timeTotal += action['value']
			self.timeCount += 1
			self.timeAvg = self.timeTotal * 1.0 / self.timeCount
			if uid in self.playUids:
				self.replayUids.add(uid)
				self.replay = len(self.replayUids)
			else:
				self.playUids.append(uid)
				self.play = len(self.playUids)
		elif actionId == 12:
			self.shareFUids.add(uid)
			self.shareF = len(self.shareFUids)
		elif actionId == 13:
			self.shareIUids.add(uid)
			self.shareI = len(self.shareIUids)
		elif actionId == 14:
			self.like += 1
		elif actionId == 15:
			self.like -= 1
		elif actionId == 16:
			self.sub += 1
		elif actionId == 17:
			self.sub -= 1
		elif actionId == 2:
			self.longPlay += 1

	def toSaveEsData(self):
		return {"vid": self.vid, "score": self.score, "time": self.endTime, "ratio": self.ratio, "aid": self.aid, "title": self.title, "play": self.play, "shareF": self.shareF, "shareI": self.shareI, "replay": self.replay, "sub": self.sub, "like": self.like, "longPlay": self.longPlay, "timeAvg": self.timeAvg, "resShow": self.resShow, "resShowPlay": self.resShowPlay}

	def updateActions(self, endTime, startTime=0):
		self.startTime = min(self.startTime, startTime) if self.startTime != -1 else startTime
		self.endTime = max(self.endTime, endTime)
		play, shareF, shareI, replay, like, sub, longPlay, timeAvg, resShow, resShowPlay = self.getDataByVid(self.vid, startTime, endTime)
		self.play += play
		self.shareF += shareF
		self.shareI += shareI
		self.replay += replay
		self.like += like
		self.sub += sub
		self.longPlay += longPlay
		self.timeAvg = (self.timeAvg + timeAvg) / 2.0 if self.timeAvg != 0 else timeAvg
		self.resShow += resShow
		self.resShowPlay += resShowPlay
		gc.collect()

	def getDataByVid(self, vid, startTime, endTime):
		#now = time.time()

		body = {"query": {"bool": {"must": [{"range": {"createTime": {"gte": startTime, "lt": endTime}}}, {"term": {"vid": vid}}, {"terms": {"action": [2, 3, 5, 12, 13, 14, 15, 16, 17]}}]}}, "size": 0, "aggs": {}}
		body['aggs']['uid_diff'] = {"terms": {"field": "action", "size": 7}, "aggs": {"count": {"cardinality": {"field": "uid"}}}}
		aggs = es.searchStatisticsAggs('works_actions', body)

		#print(time.time() - now)
		
		counts = {2: 0, 3: 0, 5:0, 12: 0, 13: 0, 14: 0, 15: 0, 16: 0, 17: 0}
		for bucket in aggs['uid_diff']['buckets']:
			counts[bucket['key']] = bucket['count']['value']

		body = {"query": {"bool": {"must": [{"range": {"createTime": {"gte": startTime, "lt": endTime}}}, {"term": {"vid": vid}}, {"term": {"action": {"value": 1}}}]}}, "size": 0, "aggs": {}}
		body['aggs']['time_avg'] = {"avg": {"field": "value"}}
		timeAvg = es.searchStatisticsAggs('works_actions', body)['time_avg']['value']

		#print(time.time() - now)

		replay = 0
		hits = es.searchStatisticsDataByBody("works_players", {"query": {"term": {"vid": vid}}, "size": 10000}, True, False, True)
		for hit in hits:
			if hit['_version'] >= 2:
				replay += 1

		#print(time.time() - now)

		#print(len(hits), counts[12], counts[13], replay, counts[16] - counts[17], counts[14] - counts[15], counts[2], timeAvg if timeAvg != None else 0)
		return len(hits), counts[12], counts[13], replay, counts[16] - counts[17], counts[14] - counts[15], counts[2], timeAvg if timeAvg != None else 0, counts[3], counts[5]


	'''
	def getDataByVid(self, vid, startTime, endTime):
		hits = es.searchStatisticsDataByBody("works_actions", {"query": {"bool": {"must": [{"range": {"createTime": {"gte": startTime, "lt": endTime}}}, {"term": {"vid": vid}}, {"terms": {"action": [1, 2, 12, 13, 14, 15, 16, 17]}}]}}}, True, False)
		playUids = []
		shareFUids = set()
		shareIUids = set()
		replayUids = set()
		like = 0
		sub = 0
		longPlay = 0
		timeTotal = 0
		timeCount = 0
		for hit in hits:
			action = hit['_source']
			actionId = action['action']
			uid = action['uid']
			if actionId == 1:
				timeTotal += action['value']
				timeCount += 1
				if uid in playUids:
					replayUids.add(uid)
				else:
					playUids.append(uid)
			elif actionId == 12:
				shareFUids.add(uid)
			elif actionId == 13:
				shareIUids.add(uid)
			elif actionId == 14:
				like += 1
			elif actionId == 15:
				like -= 1
			elif actionId == 16:
				sub += 1
			elif actionId == 17:
				sub -= 1
			elif actionId == 2:
				longPlay += 1
		return len(playUids), len(shareFUids), len(shareIUids), len(replayUids), like, sub, longPlay, timeTotal * 1.0 / timeCount if timeCount != 0 else 0
	'''


class WorksThread(threading.Thread):

	def __init__(self, works, endTime):
		threading.Thread.__init__(self)
		self.works = works
		self.endTime = endTime

	def run(self):
		self.works.updateActions(self.endTime)


